using System;
using Utils;
using System.Collections.Generic;

namespace OPSM
{
	/// <summary>
	/// 
	/// </summary>
	public class DFSMiner : Miner
	{
		public DFSMiner(Dataset ds, DualCompare dc)
            : base(ds, dc)
        {
		}

		override public void Mine(int support, int minLength, int maxLength, int maxMistakes, MineResults mineResult)
		{
			Itemset head = new Itemset();
            List<int> tail = new List<int>();

			for (int loop = 0; loop < _ds.GetColumnCount(); loop++)
				tail.Add(loop);

			RecurseMining(head, tail, support, minLength, maxLength, mineResult);
		}

        void RecurseMining(Itemset head, List<int> tail, int support, int minLength, int maxLength, MineResults mineResult)
		{
			Itemset newHead;
            List<int> newTail;

			if (head.Count >= maxLength)
				return;

			if (head.Count + tail.Count < minLength)
				return;

            List<FastSparseBitArray> bitArrays = new List<FastSparseBitArray>();
			for (int loopTail = 0; loopTail < tail.Count; loopTail++)
			{
				int i = tail[loopTail];

				if (head.Count == 0)
				{
					newTail = new List<int>(tail);
					newTail.RemoveAt(loopTail);
					newHead = new Itemset();
					newHead.AddItem(i);
					RecurseMining(newHead, newTail, support, minLength, maxLength, mineResult);
				}
				else
				{
					FastSparseBitArray bitArray = null;
					
					if (_dualComp != null)
					{
						ISimpleItemset dualItemset = _dualComp.GetItemset(head.GetLastItem(), i);
						if (dualItemset == null)
						{
							tail.RemoveAt(loopTail);
							loopTail--;
							continue;
						}

						bitArray = dualItemset.GetTransactions();
					}
					else
					{
						bitArray = FastSparseBitArrayPool.bitArrayPool.Allocate();
						_ds.BuildBitVector(head.GetLastItem(), i, bitArray);
					}

					if (head.Count > 1)
						bitArray = bitArray.And(head.GetTransactions());

					bitArray.frequency = bitArray.CountElements();
					if (bitArray.frequency >= support)
					{
						bitArrays.Add(bitArray);
					}
					else
					{
							// Don't release bit vectors from O2 matrix, or there is no
							// O2 matrix
						if ((head.Count > 1)||(_dualComp == null))
							FastSparseBitArrayPool.bitArrayPool.Release(bitArray);

						tail.RemoveAt(loopTail);
						loopTail--;
					}
				}				
			}

			if (head.Count > 0)
				for (int loopTail = 0; loopTail < tail.Count; loopTail++)
				{
					int i = tail[loopTail];

					newTail = new List<int>(tail);
					newTail.RemoveAt(loopTail);
					newHead = new Itemset(head);
					newHead.AddItem(i);
					FastSparseBitArray bitArray = bitArrays[loopTail];
					newHead.SetTransactions(bitArray);
					newHead.support = bitArray.frequency;
					RecurseMining(newHead, newTail, support, minLength, maxLength, mineResult);

					if (newHead.Count >= minLength)
						mineResult.Add(newHead);

					newHead.SetTransactions(null);
					if (newHead.Count > 2)
						FastSparseBitArrayPool.bitArrayPool.Release(bitArray);
				}
		}
	}	
}
